/*  vsa-plugin.h - header file for vsa vis and bg plugins
 *  Copyright (C) 1999  Charles <int@linuxcore.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 * $Id: vsa-plugin.h,v 1.20 1999/12/19 14:29:00 int Exp $
 */

/* If you read this that probably means you're about to write a plugin.
 * remember this stuff isn't final yet and will most likely get changed
 * frequently.  suggestions appreciated.
 */

#ifndef __VSA_PLUGIN_H
#define __VSA_PLUGIN_H

#include <stdio.h>
#include <malloc.h>
#include <string.h>

/*
 * Plugin version
 * change the major version number when there are changes that make it
 * incompatible with older versions
 */
#define VSAP_VERSION 0x0401

#define EXPORT

#ifdef __cplusplus
extern "C" {
#endif //__cplusplus
	
	
/*
 * enumerations
 */
	
typedef enum {
	VSAP_EVENT_RESET,     // reset plugin to initial state (not used..)
        VSAP_EVENT_PAINT,     // paint the image
	VSAP_EVENT_RESIZE,    // image size changed
	VSAP_EVENT_VARCHANGE, // dispatched immediately after a var is changed 

	VSAP_EVENT_BUTTON_PRESS,
	VSAP_EVENT_BUTTON_RELEASE
} VSAP_EVENT;


// the format this plugin wants the audio data in.
typedef enum {
	VSAP_FORMAT_NONE = 0, // plugin doesn't accept audio data
	VSAP_FORMAT_RAW,      // raw audio
        VSAP_FORMAT_FFT_RAW,  // raw fft output
	VSAP_FORMAT_FFT_PS,   // fft power spectrum
        VSAP_FORMAT_FFT_PCT   // scaled to width, 0-100%
} VSAP_FORMAT;



typedef enum {
	VSAP_FLAG_NONE   = 0,
	VSAP_FLAG_STEREO = 1  // plugin wants stereo audio data
} VSAP_FLAGS;



// type of the plugin
typedef enum {
	VSAP_TYPE_NONE = 0,
	VSAP_TYPE_VIS,        // visualization plugin
	VSAP_TYPE_BG,         // background plugin
	VSAP_TYPE_FILTER,     // filter plugin
	VSAP_TYPE_EXTERN,     // vis plugins that have their own window
                              // or fullscreen, OpenGL etc.
	VSAP_TYPE_MISC
} VSAP_TYPE;


typedef enum {
	VSAP_VTYPE_END = 0,   // special value to indicate end of list
	VSAP_VTYPE_BOOL,      // true or false (checkbox)
	VSAP_VTYPE_INT,       // integer (input field, opt. range)
	VSAP_VTYPE_FLOAT,     // float (input field, opt. range)
	VSAP_VTYPE_STRING,    // text (input field)

	VSAP_VTYPE_COLOR,     // VSA_Color (color picker)
	VSAP_VTYPE_LIST,      // list of strings (option menu)
	VSAP_VTYPE_MSG,

	VSAP_VTYPE_SEPARATOR  // a separator in the gui (not a type)

} VSAP_VTYPE;


typedef enum {
	VSAP_CHAN_STEREO = 0,
	VSAP_CHAN_MONO,
	VSAP_CHAN_LEFT,
	VSAP_CHAN_RIGHT
} VSAP_CHAN;



/*
 * structures
 */


// image buffer
typedef struct {
        int width;            // image width 
        int height;           // image height
        unsigned char *data;  // image RGB data
} VSA_Image;


// generic buffer for fft/raw/etc audio data
typedef struct {
	int rate;             // sample rate
	int n_chan;           // number of channels
	int n_bits;           // number of bits / sample

	int raw_size;
	short **raw;

	int size;             // data size / channel
	void **data;          // audio data, n_chan channels
			
} VSA_Audio;

// some convenience macros
#define V_AUDIO_RAW(audio, chan)        (short *)(audio)->data[chan]
#define V_AUDIO_FFT(audio, chan)        (double *)(audio)->data[chan]
#define V_AUDIO_PCT(audio, chan)        (int *)(audio)->data[chan]



// RGB color
//FIXME: should probably add alpha
typedef struct {
        unsigned char r, g, b;
} VSA_Color;


typedef struct {
	int width;
	int height;
} VSA_Size;

typedef struct {
	int x;
	int y;
} VSA_Point;

typedef struct {
	int x;
	int y;
	int width;
	int height;
} VSA_Rect;


typedef struct {
	char *target;
	char *cmd;
	void *arg;
} VSA_Msg;


/*
 * vis plugins receive this struct in paint events
 */
typedef struct {
	VSA_Image *image;
	VSA_Audio *audio;
} VSA_Data;



// vsa plugin variable
typedef union {
	int ivalue;           // integer value
	float fvalue;         // float value  
	char *svalue;         // string value
	
	VSA_Color color;      // color value
} VSA_Var;



/* plugins can define a varlist for configuration purposes.
 */
typedef struct {
	VSAP_VTYPE type;      // variable type
	char *id;             // id string
	char *parms;          // var parms, such as range
	char *desc;           // human readable description of this var
	
	void *vp;             /* optional pointer to a variable
			       * if this is NULL, get/setvar will be called
			       * use this if you don't need notification
			       * when a variable is changed
			       */
	int idx;
} VSA_VarDef;



// functions in VSA callable by plugins
typedef struct {
	void *(*send_message)(void *h, char *target, char *cmd, char *arg);
	VSA_Image *(*load_image)(void *h, char *filename);
	void (*add_var)(void *, VSAP_VTYPE, char *, char *, char *, void *);
	//void (*error_dialog)(void *h, char *msg);
} VSA_PluginFunc;


// this struct should be pretty stable now...hmm or not
typedef struct {
        int vsa_version;        // set this to VSAP_VERSION

        const char *name;       // plugin name
        const char *desc;       // plugin description
        const char *author;     // plugin author
        const char *copyright;  // plugin copyright
        const char *version;    // plugin version

        VSAP_TYPE   type;
        VSAP_FORMAT format;
        VSAP_FLAGS  flags;

	// reserved
        unsigned long r1, r2, r3, r4;
	
	void *private;

	// function table, filled in by VSA
	VSA_PluginFunc *ftab;
	void *pinfo; // pointer to VSA_PluginInfo
} VSA_Plugin;


/* function table */
static VSA_PluginFunc *vsaftab = 0;

#define INIT_PLUGIN(x, p) x->private = calloc(sizeof(p), 1); __init_plugin(x)
static __inline__ void __init_plugin(VSA_Plugin *p)
{
	//printf("__init_plugin: %p\n", p);
	p->vsa_version = VSAP_VERSION;
	vsaftab = p->ftab;
}


/*
 * The following functions can only be called from initialized plugins
 *
 */

static __inline__ void vsa_send(char *target, char *msg, void *arg)
{
	if(vsaftab && vsaftab->send_message)
		vsaftab->send_message(0, target, msg, arg);
}

static __inline__ VSA_Image *vsa_image_load(char *filename)
{
	if(vsaftab && vsaftab->load_image)
		return vsaftab->load_image(0, filename);
	return 0;
}

#define ADDVAR vsa_add_var
static __inline__ void vsa_add_var(VSA_Plugin *p, VSAP_VTYPE type,
				   char *id, char *parms,
				   char *desc, void *vp)
{
	//printf("addvar: %p  %d  %s  %s  %s  %p\n", p, type, 
	//       id, parms, desc, vp);

	if(vsaftab && vsaftab->add_var)
		vsaftab->add_var(p, type, id, parms, desc, vp);
}


/*
 * some inline util functions.
 * callable from within VSA and plugins
 * should probably move 'em to a lib if we get too many.
 */


// setup a new image buffer
static __inline__ VSA_Image *vsa_image_new(int w, int h)
{
	VSA_Image *d;

	d = (VSA_Image *)malloc(sizeof(VSA_Image));
	d->width  = w;
	d->height = h;
	d->data   = (unsigned char *)calloc(w * h,  3);

	return d;
}

// clear image
static __inline__ void vsa_image_clear(VSA_Image *d)
{
	memset(d->data, '\0', d->width * d->height * 3);
}

// free an image
static __inline__ void vsa_image_delete(VSA_Image *d)
{
	if(d->data)
		free(d->data);
	free(d);
}

// copy an image, both images must have the same size
static __inline__ int vsa_image_copy(VSA_Image *target, VSA_Image *source)
{
	if(source->width != target->width || source->height != target->height)
		return 0;
	memcpy(target->data, source->data, source->width * source->height * 3);
	return 1;
}

// copy a region of an image to another
static __inline__ int vsa_image_copy_region(VSA_Image *target, int tx, int ty,
					    VSA_Image *source, int sx, int sy,
					    int w, int h)
{
	int y;

	if(sx + w > source->width || sy + h > source->height)
		return 0;
	if(tx + w > target->width || ty + h > target->height)
		return 0;

	for(y = 0; y < h; y++)
		memcpy(&target->data[(tx + ((ty + y) * target->width)) * 3],
		       &source->data[(sx + ((sy + y) * source->width)) * 3],
		       w * 3);
	return 1;

}

//FIXME: optimize
static __inline__ int vsa_image_overlay_region(VSA_Image *target, 
					       int tx, int ty,
					       VSA_Image *source, 
					       int sx, int sy,
					       int w, int h)
{
	register int x, y;

	if(sx + w > source->width || sy + h > source->height)
		return 0;
	if(tx + w > target->width || ty + h > target->height)
		return 0;

	for(y = 0; y < h; y++)
		for(x = 0; x < w; x++)
		{
			VSA_Color *sc, *tc;

			sc = (VSA_Color *)&source->data[
				(sx + x + ((sy + y) * source->width)) * 3];
		
			if(!sc->r && !sc->g && !sc->b)
				continue;
			
			tc = (VSA_Color *)&target->data[
				(tx + x + ((ty + y) * target->width)) * 3];

			tc->r = sc->r;
			tc->g = sc->g;
			tc->b = sc->b;
		
		}
	return 1;

}


// draw a single pixel (no bounds checking)
static __inline__ void vsa_draw_pixel_noclip(VSA_Image *d, int x, int y,
					     VSA_Color *c)
{
	register unsigned char *p;

	p = d->data + ((y * d->width) + x) * 3;
	
	*p++ = c->r;
	*p++ = c->g;
	*p   = c->b;
}

// draw a single pixel
static __inline__ void vsa_draw_pixel(VSA_Image *d, int x, int y, 
				      VSA_Color *c)
{
	register unsigned char *p;

	if(x < 0 || y < 0 || x >= d->width || y >= d->height)
		return;

	p = d->data + ((y * d->width) + x) * 3;
	
	*p++ = c->r;
	*p++ = c->g;
	*p   = c->b;
}

/* Draw alpha pixel */
static __inline__ void vsa_draw_apixel(VSA_Image *d, int x, int y,
              VSA_Color *c, unsigned char alpha)
{
  register unsigned char *p;
  register unsigned char inv_a;


  if(x < 0 || y < 0 || x >= d->width || y >= d->height)
    return;

  inv_a = 255 - alpha;
  p = d->data + ((y * d->width) + x) * 3;

  *p++ = ( c->r * alpha + *p * inv_a ) / 255;
  *p++ = ( c->g * alpha + *p * inv_a ) / 255;
  *p   = ( c->b * alpha + *p * inv_a ) / 255;
}

static __inline__ int vsa_get_pixel(VSA_Image *d, int x, int y, VSA_Color *c)
{
	register unsigned char *p;

	if(x < 0 || y < 0 || x >= d->width || y >= d->height)
		return 0;

	p = d->data + ((y * d->width) + x) * 3;
	
	c->r = *p++;
	c->g = *p++;
	c->b = *p;

	return 1;
}


static __inline__ int vsa_is_pixel_clear(VSA_Image *d, int x, int y)
{
	register unsigned char *p;

	if(x < 0 || y < 0 || x >= d->width || y >= d->height)
		return 0;

	p = d->data + ((y * d->width) + x) * 3;
	
	if(!(*p) && !(*++p) && !(*++p))
		return 1;
	return 0;
}


static __inline__ void vsa_draw_vline(VSA_Image *d, int x, int y1, int y2,
				      VSA_Color *c)
{
	int y;
	if(y1 < y2)
	{
		for(y = y1; y <= y2; y++)
			vsa_draw_pixel(d, x, y, c);
	}
	else if(y2 < y1)
	{
		for(y = y2; y <= y1; y++)
			vsa_draw_pixel(d, x, y, c);
	}
	else
		vsa_draw_pixel(d, x, y1, c);
		
}


// fill a rectangular area with the specified color
static __inline__ void vsa_fill_rect(VSA_Image *d, int x, int y, 
				     int w, int h, VSA_Color *c)
{
	register int a, b;

	if(w == -1)
		w = d->width;
	if(h == -1)
		h = d->height;

	for(a = 0; a < w; a++)
		for(b = 0; b < h; b++)
			vsa_draw_pixel(d, x + a, y + b, c);
}

/* Fill a rect with an alpha color */
static __inline__ void vsa_fill_arect(VSA_Image *d, int x, int y,
            int w, int h, VSA_Color *c, unsigned char alpha)
{
	register int a, b;

	if(w == -1)
		w = d->width;
	if(h == -1)
		h = d->height;

	for(a = 0; a < w; a++)
		for(b = 0; b < h; b++)
			vsa_draw_apixel(d, x + a, y + b, c, alpha);
}



static __inline__ void vsa_fade_colors(VSA_Color c1, VSA_Color c2,
				       VSA_Color *fade, int n_col)
{
        int n;
        float step1, step2, step3;

        step1 = (c2.r - c1.r) / (float)n_col;
        step2 = (c2.g - c1.g) / (float)n_col;
        step3 = (c2.b - c1.b) / (float)n_col;

        for(n = 0; n < n_col; n++)
        {
                fade[n].r = c1.r + (step1 * n);
                fade[n].g = c1.g + (step2 * n);
                fade[n].b = c1.b + (step3 * n);
        }
}


#ifdef __cplusplus
}
#endif //__cplusplus


#endif //__VSA_PLUGIN_H
